//*****************************************************************************
// FILE:  ASSN.H
//
// DESCRIPTION:
//
//      Contains the base classes needed to implement binary associations.
//
//      This code uses the terms "left-hand-side" and "right-hand-side."
//      The left-hand-side is the point of reference - there is one and only
//      one instance on the left-hand-side.  The right-hand-side is what the
//      object "sees" as it "looks out" through the association, as if
//      the object was a sailor looking through a periscope.  In a one-to-many 
//      relationship, the left-hand-side sees many objects.  In a one-to-one 
//      relationship, the left-hand-side sees one object. 
//
// REVISION HISTORY:
//  DATE      WHO   DESCRIPTION
//  11/04/95  TJL   Created.
// 
//*****************************************************************************

#define AS_ASSN_H

//*****************************************************************************
//
// Class ASNoProperties
//
// This class is used when instantiating template classes for associations 
// that do not have properties.  This is a public class.
// 
//*****************************************************************************

class ASNoProperties
{
};

//*****************************************************************************
//
// Enum ASDeletePropertiesEnum
//
//*****************************************************************************

enum ASDeletePropertiesEnum
{  
  ASDP_DELETE_PROPERTIES, 
  ASDP_DO_NOT_DELETE_PROPERTIES 
};

//*****************************************************************************
//
// Class ASAssn
//
// This is a base class for the association subclasses.
//
//*****************************************************************************

template <class lhsClass, class rhsClass, class propClass>
class ASAssn
{
  protected:

    typedef ASAssn<rhsClass, lhsClass, propClass> rhsAssnClass;

  public:

  	// Cannot make lhsClass be const because we need to return a pointer or
    // reference to the object later.
    ASAssn( lhsClass &rLhsClass );
    virtual ~ASAssn();
    
    lhsClass &GetLhsObject();
    virtual void DetachMe() = 0;
    virtual int IsAttached() const = 0;

    int operator == ( const ASAssn<lhsClass, 
                                   rhsClass, 
                                   propClass> &rAssn 
                    ) const;
    
    //**************************************************************************
    //
    // iAttach and iDetach are internal methods.  Please do not call these 
    // directly.  
    //
    // I have tried declaring friend classes:
    //
    //  friend class ASToOneAssociation<rhsClass,lhsClass>
    //  friend class ASToOneProp<rhsClass,lhsClass,propClass>
    //
    // The first friend works but the second one creates compile errors.
    //
    //**************************************************************************

    virtual void iAttach( rhsAssnClass &,
                          propClass * const pProp = NULL,
                          const int iOrder = 0
                        ) = 0;

    virtual void iDetach( const rhsAssnClass & ) = 0;

  private:

    lhsClass *m_pLhsItem;
};

#include "assn_one.h"
#include "assnmany.h"
#include "assnimpl.h"
