//*****************************************************************************
// FILE:  ASSNIMPL.H
//
// DESCRIPTION:
//
//      Contains the implementation for all association classes.
//      This headerfile must be included in at least one .cpp file in
//      order to resolve all external symbols.
//
// REVISION HISTORY:
//  DATE      WHO   DESCRIPTION
//  11/04/95  TJL   Created.  Moved from ASSOC.H
//  11/11/95  TJL   Fix "DoNotDeleteProperties" bug.
//  11/12/95  TJL   Fix bug in ASToOneProp::iAttach.  Wasn't breaking
//                  the link.  Added tests to make sure users don't
//                  try to attach one ASAssn object to itself. 
// 
//*****************************************************************************

#define AS_ASSNIMPL_H

#ifndef AS_ASSN_H
#include "assn.h"
#endif

#ifndef AS_ASSN_ONE_H
#include "assn_one.h"
#endif

#ifndef AS_ASSNLIST_H
#include "assnlist.h"
#endif

#ifndef AS_ASSNMANY_H
#include "assnmany.h"
#endif

//*****************************************************************************
// ASAssn
//*****************************************************************************

//*****************************************************************************
//
// ASAssn::ASAssn
//
// DESCRIPTION:
//
//      Constructor.
//
// INPUT PARAMETERS:
//      rLhsClass - A reference to the object that "owns" this association.
//                  Also known as the left-hand-side object.  The right-
//                  hand-side object(s) are specified via the Attach()
//                  methods. 
//
//*****************************************************************************

template < class lhsClass, class rhsClass, class propClass >
ASAssn< lhsClass, rhsClass, propClass >::
  ASAssn< lhsClass, rhsClass, propClass >( lhsClass &rLhsClass ) 
  : m_pLhsItem( &rLhsClass )
{
}
    
//*****************************************************************************
//
// ASAssn::~ASAssn
//
// DESCRIPTION:
//
//      Destructor.
//
//*****************************************************************************
template < class lhsClass, class rhsClass, class propClass >
ASAssn< lhsClass, rhsClass, propClass >::
  ~ASAssn< lhsClass, rhsClass, propClass >()
{
}
    
//*****************************************************************************
//
// ASAssn::GetLhsObject
//
// DESCRIPTION:
//
//      Returns a reference to the "left-hand-side" object that is
//      involved in the association.  The left-hand-side object is specified
//      at construction-time and remains the "left-hand-side" object
//      for the life of "this". 
//
//*****************************************************************************

template < class lhsClass, class rhsClass, class propClass >
lhsClass &
ASAssn< lhsClass, rhsClass, propClass >::GetLhsObject()
{
  return *m_pLhsItem;
}
    
//*****************************************************************************
//
// ASAssn::operator =
//
// DESCRIPTION:
//
//      Tests whether two ASAssn objects are equal.  Equality
//      is based on object identity; that is, their addresses.  This 
//      method is called by list classes that are involved in one-to-many 
//      relationships; particularly, the RemoveItem class.
//
// INPUT PARAMETERS:
//      rAssn - A reference to an association object.
//
// RETURNS:
//      0 - The association objects are not equal
//   Else - The association objects are identical
//
//*****************************************************************************

template < class lhsClass, class rhsClass, class propClass >
int
ASAssn< lhsClass, rhsClass, propClass >::operator ==
  ( const ASAssn<lhsClass, rhsClass, propClass> &rAssn ) const
{
  return this == &rAssn;
}
    
//*****************************************************************************
// ASToOneBase
//*****************************************************************************

//*****************************************************************************
//
// ASToOneBase::ASToOneBase
//
// DESCRIPTION:
//
//      Constructor.
//
// INPUT PARAMETERS:
//      rItem - A reference to the left-hand-side object.
//
//*****************************************************************************

template < class lhsClass, class rhsClass, class propClass >
ASToOneBase<lhsClass, rhsClass, propClass>::
  ASToOneBase<lhsClass, rhsClass, propClass>( lhsClass & rItem )  
  : m_pRhsAssociation(NULL),
    ASAssn< lhsClass, rhsClass, propClass > ( rItem )
{
}

//*****************************************************************************
//
// ASToOneBase::IsAttached
//
// DESCRIPTION:
//
//      Call this to determine whether the left-hand-side object is linked
//      to another object.
//
// RETURNS:
//      0 - The left-hand-side object is not linked.
//      1 - The left-hand-side object is linked.
//
//*****************************************************************************

template <class lhsClass, class rhsClass, class propClass>
int 
ASToOneBase<lhsClass, rhsClass, propClass>::IsAttached() const
{
  return m_pRhsAssociation ? 1 : 0;
}

//*****************************************************************************
//
// ASToOneBase::GetRhsObject
//
// DESCRIPTION:
//
//      Provides access to the right-hand-side object in the association.
//      There must be a right-hand-side object.
//
// RETURNS:
//      A reference to the object.
//
// EXCEPTIONS:
//      0 - If there is no right-hand-side object. 
//
//*****************************************************************************

template <class lhsClass, class rhsClass, class propClass>
rhsClass &
ASToOneBase<lhsClass, rhsClass, propClass>::
  GetRhsObject()
{
  if ( !m_pRhsAssociation )
  {
    throw 0;
  }
  return m_pRhsAssociation -> GetLhsObject();
}    

//*****************************************************************************
// ASToOne
//*****************************************************************************

//*****************************************************************************
//
// ASToOne::ASToOne
//
// DESCRIPTION:
//
//      Constructor.
//
// INPUT PARAMETERS:
//      rItem - A reference to the left-hand-side object.
//
//*****************************************************************************

template< class lhsClass, class rhsClass >
ASToOne< lhsClass, rhsClass >::
  ASToOne< lhsClass, rhsClass >( lhsClass & rItem )  
  : ASToOneBase< lhsClass, rhsClass, ASNoProperties > ( rItem )
{
}
    
//*****************************************************************************
//
// ASToOne::~ASToOne
//
// DESCRIPTION:
//
//      Destructor.  Detaches *this from its associated object. 
//
//      The destructor must be implemented on this class because the
//      overloaded DeleteMe method cannot be called by the parent
//      class's destructor.
//
//*****************************************************************************

template< class lhsClass, class rhsClass >
ASToOne< lhsClass, rhsClass >::
  ~ASToOne< lhsClass, rhsClass >()
{
  DetachMe();
}
    
//*****************************************************************************
//
// ASToOne::DetachMe
//
// DESCRIPTION:
//
//      Detaches the left-hand-side and right-hand-side objects.  They
//      will no longer be associated with each other.
//
//      If there is no right-hand-side object, nothing happens.
//
//*****************************************************************************

template< class lhsClass, class rhsClass >
void
ASToOne< lhsClass, rhsClass >::DetachMe()
{
  if ( m_pRhsAssociation )
  {
    m_pRhsAssociation -> iDetach( *this );
    m_pRhsAssociation = NULL;
  }
}   
    
//*****************************************************************************
//
// ASToOne::Attach
//
// DESCRIPTION:
//
//      Links two objects together.  Both objects will "point" to 
//      each other, via their ASAssociation objects. 
//
// INPUT PARAMETERS:
//      prAssociationToLinkTo - The association object to link to.
//
// EXCEPTIONS:
//      2 - rAssociationToLinkTo is the same as "this"
//
//*****************************************************************************

template< class lhsClass, class rhsClass >
void
ASToOne< lhsClass, rhsClass >::Attach( rhsAssnClass & rAssociationToLinkTo )
{
  if ( this == (ASToOne<lhsClass, rhsClass> *) &rAssociationToLinkTo )
  {
    throw 2;
  }
  if ( m_pRhsAssociation )
  {
    m_pRhsAssociation -> iDetach( *this );
  }
  rAssociationToLinkTo.iAttach( *this );
  m_pRhsAssociation = &rAssociationToLinkTo;
}
    
//*****************************************************************************
//
// ASToOne::iAttach
//
// DESCRIPTION:
//
//      Attaches two objects.  This method differs from Attach() in that 
//      the other association object is not sent a corresponding iAttach() 
//      message.
//
// INPUT PARAMETERS:
//      rAssociation - The other association object to link to.
//      pProperties - Ignored. 
//      iOrder - Ignored.
//
//*****************************************************************************

template< class lhsClass, class rhsClass >
void
ASToOne< lhsClass, rhsClass >::
  iAttach( rhsAssnClass & rAssociation,
           ASNoProperties * const,
           const int iOrder
         )
{
  if ( m_pRhsAssociation )
  {
    m_pRhsAssociation -> iDetach( *this );
  }
  m_pRhsAssociation = &rAssociation;
}
    
//*****************************************************************************
//
// ASToOne::iDetach
//
// DESCRIPTION:
//
//      Internal detach method.  Detaches *this from another
//      association.  This function differs from Detach() in
//      that it doesn't send a corresponding iDetach() message to the
//      other association object.
//
// INPUT PARAMETERS:
//      rAssociation - The association to detach the left-hand-side object
//                     from. 
//
//*****************************************************************************

template< class lhsClass, class rhsClass >
void 
ASToOne< lhsClass, rhsClass >::
  iDetach( const rhsAssnClass &rAssociation )
{
  m_pRhsAssociation = NULL;
}
    
//*****************************************************************************
// ASToOneProp
//*****************************************************************************

//*****************************************************************************
//
// ASToOneProp::ASToOneProp
//
// DESCRIPTION:
//
//      Constructor.
//
// INPUT PARAMETERS:
//      rItem - The left-hand-side object.
//      eDeleteProperties - Specifies whether or not to delete the property
//                          object when the association is revoked.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
ASToOneProp< lhsClass, rhsClass, propClass >::
  ASToOneProp< lhsClass, rhsClass, propClass >
  ( lhsClass & rItem, 
    const ASDeletePropertiesEnum eDeleteProperties
  )
  : ASToOneBase< lhsClass, rhsClass, propClass > ( rItem ),
    m_eDeleteProperties( eDeleteProperties )
{
}

//*****************************************************************************
//
// ASToOneProp::~ASToOneProp
//
// DESCRIPTION:
//
//      Destructor.  Detaches *this from its associated object. 
//
//      The destructor must be implemented on this class because the
//      overloaded DeleteMe method cannot be called by the parent
//      class's destructor.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
ASToOneProp< lhsClass, rhsClass, propClass >::
  ~ASToOneProp< lhsClass, rhsClass, propClass >()
{
  DetachMe();
}

//*****************************************************************************
//
// ASToOneProp::Attach
//
// DESCRIPTION:
//
//      Links two objects together.  Both objects will "point" to 
//      each other, via their ASAssociation objects. 
//
// INPUT PARAMETERS:
//      rAssociationToLinkTo - The association to link to. 
//      rProperties - The association properties.  This object may be deleted
//                    when the association is revoked -- this depends on how
//                    *this was constructed.  Both sides of an association
//                    point to the same property object.
//
// EXCEPTIONS:
//      3 - rAssociationToLinkTo is the same as "this"
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
void
ASToOneProp< lhsClass, rhsClass, propClass >::
  Attach( rhsAssnClass & rAssociationToLinkTo,
          propClass & rProperties
        )
{
  if ( this == (ASToOneProp<lhsClass, rhsClass, propClass> *) &rAssociationToLinkTo )
  {
    throw 3;
  }
  if ( m_pRhsAssociation )
  {
    m_pRhsAssociation -> iDetach( *this );
  }
  rAssociationToLinkTo.iAttach( *this, &rProperties );
  m_pRhsAssociation = &rAssociationToLinkTo;
  m_pProperties = &rProperties;
}

//*****************************************************************************
//
// ASToOneProp::GetProperties
//
// DESCRIPTION:
//
//      Returns the properties for the association.
//
// RETURNS:
//      A reference to a property object. 
//
// EXCEPTIONS:
//      1 - There are no properties, or the association is not linked with
//          another association object.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
propClass&
ASToOneProp< lhsClass, rhsClass, propClass >::GetProperties()
{
  if ( !m_pProperties )
  {
    throw 1;
  }
  return *m_pProperties;
}

//*****************************************************************************
//
// ASToOneProp::DetachMe
//
// DESCRIPTION:
//
//      Detaches the left-hand-side and right-hand-side objects.  They
//      will no longer be associated with each other.
//
//      If there is no right-hand-side object, nothing happens.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
void
ASToOneProp< lhsClass, rhsClass, propClass >::DetachMe()
{
  // Do not try to combine the DetachMe's into the base class.  This
  // will cause iDetach() to be called against *this, which will cause
  // the properties pointer to be deleted twice.
  if ( m_pRhsAssociation )
  {
    m_pRhsAssociation -> iDetach( *this );
    m_pRhsAssociation = NULL;
    m_pProperties = NULL;
  }
}   

//*****************************************************************************
//
// ASToOneProp::iAttach
//
// DESCRIPTION:
//
//      Attaches two objects.  This method differs from Attach() in that 
//      the other association object is not sent a corresponding iAttach() 
//      message.
//
// INPUT PARAMETERS:
//      rAssociation - The other association object to link to.
//      pProperties - A pointer to a property class.
//      iOrder - Ignored
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
void
ASToOneProp< lhsClass, rhsClass, propClass >::
  iAttach( rhsAssnClass & rAssociation,
           propClass * const pProperties,
           const int iOrder
         )
{
  if ( m_pRhsAssociation )
  {
    m_pRhsAssociation -> iDetach( *this );
  }
  m_pRhsAssociation = &rAssociation;
  m_pProperties = pProperties;
}

//*****************************************************************************
//
// ASToOneProp::iDetach
//
// DESCRIPTION:
//
//      Internal detach method.  Detaches *this from another
//      association.  This function differs from Detach() in
//      that it doesn't send a corresponding iDetach() message to the
//      other association object.
//
// INPUT PARAMETERS:
//      rAssociation - The association to detach the left-hand-side object
//                     from. 
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
void
ASToOneProp< lhsClass, rhsClass, propClass >::iDetach( const rhsAssnClass & )
{
  if ( m_eDeleteProperties == ASDP_DELETE_PROPERTIES )
  {
    delete m_pProperties;
  }
  m_pRhsAssociation = NULL;
  m_pProperties = NULL;
}
    
//*****************************************************************************
// ASAssnPropTuple
//*****************************************************************************

//*****************************************************************************
//
// ASAssnPropTuple::ASAssnPropTuple
//
// DESCRIPTION:
//
//      Constructor.
//
// INPUT PARAMETERS:
//      rAssociation - An association object.
//      rProperties - An "association properties" object.
//      eDeleteProperties - Specifies whether to delete the "association
//                          properties" object when the tuple object is
//                          destroyed.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
ASAssnPropTuple< lhsClass, rhsClass, propClass >::
  ASAssnPropTuple< lhsClass, rhsClass, propClass >
  ( ASAssn<lhsClass, rhsClass, propClass>&rAssociation,
    propClass &rProperties,
    const ASDeletePropertiesEnum eDeleteProperties 
  )
  : m_eDeleteProperties(eDeleteProperties),
    m_rProperties( rProperties ),
    m_rAssociation( rAssociation )
{
}

//*****************************************************************************
//
// ASAssnPropTuple::~ASAssnPropTuple
// 
//
// DESCRIPTION:
//
//      Destroys the tuple object.  If, at construction time, the
//      tuple was instructed to delete the properties, then the property
//      object will be deleted by this method.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
ASAssnPropTuple< lhsClass, rhsClass, propClass >::
  ~ASAssnPropTuple< lhsClass, rhsClass, propClass > ()
{
  if ( m_eDeleteProperties == ASDP_DELETE_PROPERTIES )
  {
    delete &m_rProperties; // Delete NULL OK in C++
  }
}

//*****************************************************************************
//
// ASAssnPropTuple::DoNotDeleteProperties
//
// DESCRIPTION:
//
//      When the tuple object is destroyed, it may delete its corresponding
//      property object, depending on m_eDeleteProperties.  Calling this
//      function prevents this behavior.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
void
ASAssnPropTuple< lhsClass, rhsClass, propClass >::DoNotDeleteProperties()
{
  m_eDeleteProperties = ASDP_DO_NOT_DELETE_PROPERTIES;
}

//*****************************************************************************
//
// ASAssnPropTuple::operator ==
// 
//
// DESCRIPTION:
//
//      Tests equality between a tuple and an association.  This function 
//      is called by a list class to check to see if it has found what
//      it is looking for, particularly in RemoveItem().
//
// INPUT PARAMETERS:
//      rAssociation - An association object (see return value description).
//
// RETURNS:
//      0 : The tuple does not point to the given association. 
//   Else : The tuple does point to the given association.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
int
ASAssnPropTuple< lhsClass, rhsClass, propClass >::
  operator == (const ASAssn<lhsClass, rhsClass, propClass> &rAssociation) const
{
  return &m_rAssociation == &rAssociation;
}

//*****************************************************************************
//
// ASAssnPropTuple::GetAssn
//
// DESCRIPTION:
//
//      Provides access to the association object. 
//
// RETURNS:
//      Returns a (non-const) reference to the association object.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
ASAssn<lhsClass, rhsClass, propClass> & 
ASAssnPropTuple< lhsClass, rhsClass, propClass >::GetAssn()
{
  return m_rAssociation;
}

//*****************************************************************************
//
// ASAssnPropTuple::GetProperties
//
// DESCRIPTION:
//
//      Provides access to the association properties. 
//
// RETURNS:
//      A reference to a property object. 
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass >
propClass &
ASAssnPropTuple< lhsClass, rhsClass, propClass >::GetProperties()
{
  return m_rProperties;
}
    
//*****************************************************************************
// ASToManyBase
//*****************************************************************************

//*****************************************************************************
//
// ASToManyBase::ASToManyBase
//
// DESCRIPTION:
//
//      Constructor.
//
// INPUT PARAMETERS:
//      rItem - The left-hand-side object.
//      rList - A reference to a list object that has the ASListIterator
//              interface.  This object should be allocated on the heap,
//              via "new".  It is destroyed along with *this.  This list 
//              is used to store the "pointers" to the objects that are 
//              associated with rItem.
//
//              Note that the name of the concrete list class is not a template 
//              parameter.  ASToManyBase doesn't need to know specifics (and
//              therefore it shouldn't know specifics); it just needs to know 
//              about the interface.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass, class listItemClass >
ASToManyBase< lhsClass, rhsClass, propClass, listItemClass >::
  ASToManyBase< lhsClass, rhsClass, propClass, listItemClass >
  ( lhsClass & rItem, listAdapterClass &rList )  
  : ASAssn< lhsClass, rhsClass, propClass > ( rItem ),
    m_rAssnList(rList)
{
}

//*****************************************************************************
//
// ASToManyBase::~ASToManyBase
//
// DESCRIPTION:
//
//      Destructor.  Destroys the list that was passed in.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass, class listItemClass >
ASToManyBase< lhsClass, rhsClass, propClass, listItemClass >::
  ~ASToManyBase<lhsClass, rhsClass, propClass, listItemClass> ()
{
  delete &m_rAssnList;
}

//*****************************************************************************
//
// ASToManyBase::IsAttached
//
// DESCRIPTION:
//
//      Determines whether the left-hand-side object is associated with
//      other objects.
//
// RETURNS:
//      0 - No
//      1 - Yes
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass, class listItemClass >
int
ASToManyBase< lhsClass, rhsClass, propClass, listItemClass >::
  IsAttached() const
{
  return m_rAssnList.GetNumItems() > 0;
}

//*****************************************************************************
//
// ASToManyBase::GetRhsObjects
//
// DESCRIPTION:
//
//      Provides access to the associated objects.
//
// RETURNS:
//      Returns a list that can be traversed to find the objects that
//      the left-hand-side object is associated with.
//
//      If the association has properties, then there are ASAssnPropTuple
//      objects in the list.  Otherwise, the list contains ASAssn objects.
//
//*****************************************************************************

template< class lhsClass, class rhsClass, class propClass, class listItemClass >
const ASToManyBase<lhsClass, rhsClass, propClass, listItemClass>::listAdapterClass &
ASToManyBase< lhsClass, rhsClass, propClass, listItemClass >::
  GetRhsObjects() const
{
  return m_rAssnList;
}
    
//*****************************************************************************
// ASToMany
//*****************************************************************************

//*****************************************************************************
//
// ASToMany::ASToMany
//
// DESCRIPTION:
//
//      Constructor.
//
// INPUT PARAMETERS:
//      param1 - Description
//      param2 - Description
//
//*****************************************************************************

template < class lhsClass, class rhsClass >
ASToMany< lhsClass, rhsClass >::
  ASToMany< lhsClass, rhsClass >
  ( lhsClass &rItem, listAdapterClass &rList )  
  : ASToManyBase< lhsClass,  
                  rhsClass, 
                  ASNoProperties,
                  rhsAssnClass
                > ( rItem, rList )
{
}

//*****************************************************************************
//
// ASToMany::~ASToMany
//
// DESCRIPTION:
//
//      Destructor.  Detaches *this from all other associated objects. 
//
//      The destructor must be implemented on this class because the
//      overloaded DeleteMe method cannot be called by the parent
//      class's destructor.
//
//*****************************************************************************

template < class lhsClass, class rhsClass >
ASToMany< lhsClass, rhsClass >::
  ~ASToMany< lhsClass, rhsClass >()
{
  DetachMe();
}

//*****************************************************************************
//
// ASToMany::Attach
//
// DESCRIPTION:
//
//      Attaches two objects.  Both objects will point to each other.
//
// INPUT PARAMETERS:
//      rAssociationToLinkTo - The other association object to link to.
//      iOrder - For ordered associations, this number specifies the
//               relative order that rAssociation will appear when iterating
//               through the association list. 
//
// EXCEPTIONS:
//      4 - rAssociationToLinkTo is the same as "this"
//
//*****************************************************************************

template < class lhsClass, class rhsClass >
void
ASToMany< lhsClass, rhsClass >::
  Attach( rhsAssnClass & rAssociationToLinkTo,
          int iOrder
         )
{
  if ( this == (ASToMany<lhsClass, rhsClass> *) &rAssociationToLinkTo )
  {
    throw 4;
  }
  // Am I already attached to this association? 
  if ( m_rAssnList.FindItem( rAssociationToLinkTo ) )
  {
    return;
  }
  
  rAssociationToLinkTo.iAttach( *this, NULL, iOrder );
  m_rAssnList.AddItem( rAssociationToLinkTo, iOrder );
}

//*****************************************************************************
//
// ASToMany::DetachMe
//
// DESCRIPTION:
//
//      Detaches the left-hand-side object from all other associated objects. 
//
//*****************************************************************************

template < class lhsClass, class rhsClass >
void
ASToMany< lhsClass, rhsClass >::DetachMe()
{ 
  iteratorAdapterClass * pIterator = m_rAssnList.Iterate();
  
  for ( ; !pIterator -> AtEnd(); pIterator -> Next() )
  {
    // Tell each item in the list to detach itself from *this
    pIterator -> GetCurrent().iDetach(*this);
  }
  delete pIterator;
  m_rAssnList.RemoveAll();
}

//*****************************************************************************
//
// ASToMany::Detach
//
// DESCRIPTION:
//
//      Detaches two objects.
//
// INPUT PARAMETERS:
//      rAssociation - This is the other association.  Presumably, *this
//                     is currently attached to this association.
//
//*****************************************************************************

template < class lhsClass, class rhsClass >
void
ASToMany< lhsClass, rhsClass >::
  Detach( rhsAssnClass &rAssociation )
{
  if ( !m_rAssnList.FindItem( rAssociation ) )
  {
    return;
  }

  rAssociation.iDetach(*this);
  m_rAssnList.RemoveItem( rAssociation );
}

//*****************************************************************************
//
// ASToMany::iAttach 
//
// DESCRIPTION:
//
//      Attaches two objects.  This method differs from Attach() in that 
//      the other association object is not sent a corresponding iAttach() 
//      message.
//
// INPUT PARAMETERS:
//      rAssociation - The other association object to link to.
//      pProperties - Ignored.
//      iOrder - For ordered associations, this number specifies the
//               relative order that rAssociation will appear when iterating
//               through the association list. 
//
//*****************************************************************************

template < class lhsClass, class rhsClass >
void
ASToMany< lhsClass, rhsClass >::
  iAttach( rhsAssnClass &rAssociation,
           ASNoProperties * const pProperties,
           const int iOrder
          ) 
{
  m_rAssnList.AddItem( rAssociation, iOrder );
}

//*****************************************************************************
//
// ASToMany::iDetach
//
// DESCRIPTION:
//
//      Internal detach method.  Detaches *this from another
//      association.  This function differs from Detach() in
//      that it doesn't send a corresponding iDetach() message to the
//      other association object.
//
// INPUT PARAMETERS:
//      rAssociation - The association to detach the left-hand-side object
//                     from. 
//
//*****************************************************************************

template < class lhsClass, class rhsClass >
void
ASToMany< lhsClass, rhsClass >::
  iDetach( const rhsAssnClass &rAssociation )
{
  m_rAssnList.RemoveItem( rAssociation );
}

//*****************************************************************************
// ASToManyProp
//*****************************************************************************

//*****************************************************************************
//
// ASToManyProp::ASToManyProp 
//
// DESCRIPTION:
//
//      Constructor for one-to-many association class that has
//      properties.
//
// INPUT PARAMETERS:
//      rItem - A reference to the left-hand-side object.
//      rList - A reference to an object that has the same interface as
//              ASListAdapter.
//      eDeleteProperties - Specifies whether or not to delete the property
//                          object after as association is revoked.
//
//*****************************************************************************

template <class lhsClass, 
          class rhsClass, 
          class propClass
         >
ASToManyProp< lhsClass, rhsClass, propClass >::
  ASToManyProp< lhsClass, rhsClass, propClass >
  ( lhsClass & rItem, 
    listAdapterClass & rList,  
    const ASDeletePropertiesEnum eDeleteProperties
  )
  : ASToManyBase< lhsClass, 
                  rhsClass,  
                  propClass,
                  listItemClass
                > ( rItem, rList ),
    m_eDeleteProperties( eDeleteProperties ) 
{
}

//*****************************************************************************
//
// ASToManyProp::~ASToManyProp 
//
// DESCRIPTION:
//
//      Destructor.  Detaches *this from all other associated objects. 
//
//      The destructor must be implemented on this class because the
//      overloaded DeleteMe method cannot be called by the parent
//      class's destructor.
//
//*****************************************************************************

template <class lhsClass, 
          class rhsClass, 
          class propClass
         >
ASToManyProp< lhsClass, rhsClass, propClass >::
  ~ASToManyProp< lhsClass, rhsClass, propClass > ()
{
  DetachMe();
}

//*****************************************************************************
//
// ASToManyProp::Attach 
//
// DESCRIPTION:
//
//      Attaches two objects.  Both objects will point to each other.
//
// INPUT PARAMETERS:
//      rAssociationToLinkTo - The other association object to link to.
//      rProperties - A reference to an object that contains the
//                    properties for the association.  This object may
//                    be deleted when the association is revoked -
//                    it depends on how *this was constructed.
//      iOrder - For ordered associations, this number specifies the
//               relative order that rAssociation will appear when iterating
//               through the association list. 
//
// EXCEPTIONS:
//      5 - rAssociationToLinkTo is the same as "this"
//
//*****************************************************************************

template <class lhsClass, 
          class rhsClass, 
          class propClass
         >
void
ASToManyProp< lhsClass, rhsClass, propClass >::
  Attach( rhsAssnClass & rAssociationToLinkTo,
          propClass &rProperties,
          int iOrder
        )
{
  if ( this == (ASToManyProp<lhsClass, rhsClass, propClass> *)
               &rAssociationToLinkTo 
     )
  {
    throw 5;
  }
  // Am I already attached to this association? 
  if ( m_rAssnList.FindItem( rAssociationToLinkTo ) )
  {
    return;
  }
  
  rAssociationToLinkTo.iAttach( *this, &rProperties, iOrder );
  iAttach( rAssociationToLinkTo, &rProperties, iOrder );
}

//*****************************************************************************
//
// ASToManyProp::DetachMe 
//
// DESCRIPTION:
//
//      Detaches *this from all of the objects that it is associated with. 
//      Cleans up memory by deleting all of the tuples in the list.
//
//*****************************************************************************

template <class lhsClass, 
          class rhsClass, 
          class propClass
         >
void
ASToManyProp< lhsClass, rhsClass, propClass >::DetachMe()
{ 
  iteratorAdapterClass * pIterator = m_rAssnList.Iterate();
  listItemClass *pItem;
  
  for ( ; !pIterator -> AtEnd(); pIterator -> Next() )
  {
    // Tell each item in the list to detach itself from *this
    pItem = &pIterator -> GetCurrent();
    pItem -> GetAssn().iDetach(*this);
    pItem -> DoNotDeleteProperties();
    delete pItem;
  }
  delete pIterator;
  m_rAssnList.RemoveAll();
}

//*****************************************************************************
//
// ASToManyProp::Detach 
//
// DESCRIPTION:
//
//      Detaches two objects.
//
// INPUT PARAMETERS:
//      rAssociation - This is the other association.  Presumably, *this
//                     is currently attached to this association.
//
//*****************************************************************************

template <class lhsClass, 
          class rhsClass, 
          class propClass
         >
void
ASToManyProp< lhsClass, rhsClass, propClass >::
  Detach( rhsAssnClass &rAssociation )
{
  // Are we attached to this association? 
  listItemClass *pItem = m_rAssnList.FindItem( rAssociation );
    
  if ( !pItem )
  {
    return;
  }
  rAssociation.iDetach(*this);
  m_rAssnList.RemoveItem( rAssociation );
  pItem -> DoNotDeleteProperties();
  delete pItem;
}

//*****************************************************************************
//
// ASToManyProp::iAttach 
//
// DESCRIPTION:
//
//      Attaches two objects.  This method differs from Attach() in that 
//      the other association object is not sent a corresponding iAttach() 
//      message.
//
// INPUT PARAMETERS:
//      rAssociation - The other association object to link to.
//      pProperties - A pointer to a property class.
//      iOrder - For ordered associations, this number specifies the
//               relative order that rAssociation will appear when iterating
//               through the association list. 
//
//*****************************************************************************

template <class lhsClass, 
          class rhsClass, 
          class propClass
         >
void
ASToManyProp< lhsClass, rhsClass, propClass >::
  iAttach( rhsAssnClass &rAssociation,
           propClass * const pProperties,
           const int iOrder
         ) 
{
  listItemClass *pItem = new listItemClass(rAssociation, 
                                           *pProperties,  
                                           m_eDeleteProperties
                                          );
  m_rAssnList.AddItem( *pItem, iOrder );
}

//*****************************************************************************
//
// ASToManyProp::iDetach 
//
// DESCRIPTION:
//
//      Internal detach method.  Detaches *this from another
//      association.  This function differs from Detach() in
//      that it doesn't send a corresponding iDetach() message to the
//      other association object.
//
// INPUT PARAMETERS:
//      rAssociation - The association to detach the left-hand-side object
//                     from. 
//
//*****************************************************************************

template <class lhsClass, 
          class rhsClass, 
          class propClass
         >
void
ASToManyProp< lhsClass, rhsClass, propClass >::
 iDetach( const rhsAssnClass &rAssociation )
{
  listItemClass * pItem = m_rAssnList.FindItem( rAssociation );
    
  m_rAssnList.RemoveItem( rAssociation );
  delete pItem;
}
