//*****************************************************************************
//
// FILE:  ASSNLIST.H
//
// DESCRIPTION:
//
//      Contains definitions for adapter classes used by the "one-to-many"
//      association classes.  Note: the methods are documented at the bottom
//      of this file.
//
// REVISION HISTORY:
//  DATE      WHO   DESCRIPTION
//  11/03/95  TJL   Created.
//  11/05/95  TJL   Comment the methods.
//  11/12/95  TJL   Rename helper macros.
// 
//*****************************************************************************

#define AS_ASSNLIST_H

//*****************************************************************************
//
// Class IteratorAdapter
//
// This is an abstract class that defines the interface for a list iterator.
//
//*****************************************************************************

template <class itemClass>
class ASIteratorAdapter
{
  public:
  
    virtual ~ASIteratorAdapter()
    {
      // Cannot have pure virtual destructor
    }
    
    // Determine whether the iterator has reached the end of the list
    virtual int AtEnd() = 0;
    
    // Position to the first or next item
    virtual void Next() = 0;
    
    // Get the current item
    virtual itemClass &GetCurrent() = 0; // No const version
};

//*****************************************************************************
//
// Class ASListAdapter
//
// This is an abstract class that defines the interface for a list.
// When the list is destroyed, all items in the list should be
// deleted.
//
//*****************************************************************************

template <class itemClass, class searchClass>
class ASListAdapter
{
  public:
  
    virtual ~ASListAdapter()
    {
      // Cannot have pure virtual destructor
    }

    // Adding / Insertion
    virtual void AddItem(itemClass &, int order = 0) = 0;
    
    // Removal
    virtual void RemoveItem(const searchClass&) = 0;
    virtual void RemoveAll() = 0;
    
    // Search
    virtual itemClass *FindItem(const searchClass&) = 0;
    
    // Number of Items
    virtual int GetNumItems() const = 0;
    
    // Iteration
    // You can change the items in a const list by iterating
    // through it.  This is OK. Users should not be
    // able to delete or add items from const lists. 
    virtual ASIteratorAdapter<itemClass> *Iterate() const = 0;
};

//*****************************************************************************
//
// Short-cuts for declaring concrete lists and abstract iterator adapters.
//
// The left-hand-side is the point of reference.  The right-hand-side 
// is the "many" side.
//
//*****************************************************************************

//*****************************************************************************
// AS_LIST : For declaring a concrete list.
//           Takes the name of concrete list class, the left-hand-side
//           class name, and the right-hand-side-class name.
//*****************************************************************************
#define AS_LIST( list_class, lhs_class, rhs_class ) \
  list_class< ASAssn< rhs_class, lhs_class, ASNoProperties >, \
              ASAssn< rhs_class, lhs_class, ASNoProperties > \
            >
            
//*****************************************************************************
// AS_LIST_PROP : For declaring a concrete list.
//                Takes the name of concrete list class, the left-hand-side
//                class name, the right-hand-side-class name, and 
//                the property class name.
//*****************************************************************************
#define AS_LIST_PROP( list_class, lhs_class, rhs_class, prop_class ) \
  list_class< ASAssnPropTuple< rhs_class, lhs_class, prop_class >, \
              ASAssn< rhs_class, lhs_class, prop_class > \
            >
            
//*****************************************************************************
// AS_ITER_ADAPTER : For declaring an abstract list iterator.
//                     Takes the left-hand-side class name and the 
//                     right-hand-side-class name.
//*****************************************************************************
#define AS_ITER_ADAPTER( lhs_class, rhs_class ) \
    ASIteratorAdapter< ASAssnPropTuple< rhs_class, lhs_class, ASNoProperties > \
                     >
                     
//*****************************************************************************
// AS_ITER_ADAPTER_PROP : For declaring an abstract list iterator.
//                        Takes the left-hand-side class name, the 
//                        right-hand-side-class name, and the name of the
//                        property class.
//*****************************************************************************
#define AS_ITER_ADAPTER_PROP( lhs_class, rhs_class, prop_class ) \
    ASIteratorAdapter< ASAssnPropTuple< rhs_class, lhs_class, prop_class > \
                     >
                     
//*****************************************************************************
//
// ASIteratorAdapter::Next
//
// DESCRIPTION:
//
//      Positions to the next item in the list.
//
// RETURNS:
//      0 : Reached the end of the list 
//      1 : Have not reached the end of the list yet 
//
//*****************************************************************************

//*****************************************************************************
//
// ASListAdapter::GetCurrent
//
// DESCRIPTION:
//
//      Gets the current list item.
//
// RETURNS:
//      A reference to the list item.
//
// EXCEPTIONS:
//      Throws an exception if the end of the list has been reached,
//      or if Next() was never called.
//
//*****************************************************************************

//*****************************************************************************
//
// ASListAdapter::AddItem
//
// DESCRIPTION:
//
//      Adds an item to the list.
//
// INPUT PARAMETERS:
//      rItem - A reference to the item to store in the list.  A pointer
//              to the item is stored in the list.
//      iOrder - The relative order in which to insert the item into the
//               list.  This parameter can be ignored, depending on the
//               list implementation. 
//
//*****************************************************************************

//*****************************************************************************
//
// ASListAdapter::RemoveItem
//
// DESCRIPTION:
//
//      Removes an item from the list.
//
// INPUT PARAMETERS:
//      rItem - The item to remove.
//
// EXCEPTIONS:
//      No exception is thrown, even if the item wasn't found in the list.
//
//*****************************************************************************

//*****************************************************************************
//
// ASListAdapter::RemoveAll
//
// DESCRIPTION:
//
//      Removes all entries from the list.
//
//*****************************************************************************

//*****************************************************************************
//
// ASListAdapter::FindItem
//
// DESCRIPTION:
//
//      Searches for an item in the list.
//
// INPUT PARAMETERS:
//      rItem - The item to search for.
//
// RETURNS:
//      A pointer to the item, or NULL if not found. 
//
//*****************************************************************************

//*****************************************************************************
//
// ASListAdapter::GetNumItems
//
// DESCRIPTION:
//
//      Returns the number of items currently in the list.
//
//*****************************************************************************

//*****************************************************************************
//
// ASListAdapter::Iterate
//
// DESCRIPTION:
//
//      Starts an iterator, so the items in the list can be traversed.
//      Next() must be sent to the iterator before calling GetCurrent().
//
// RETURNS:
//      A pointer to a new ASListAdapter object.  This pointer must be
//      deleted by the caller.
//
//*****************************************************************************
