//*****************************************************************************
// FILE:  TEST.CPP
//
// DESCRIPTION:
//
//      Test program for associations.
//
// REVISION HISTORY:
//  DATE      WHO   DESCRIPTION
//  10/20/95  TJL   Created.
// 
//*****************************************************************************

#include <stdio.h>
#include "assn.h"

#include "list.h"

class A
{
  public:
  int i;
};

class B
{
  public:
  int i;
};

class Properties
{
  public:
  int i;
  
  Properties( int x ) : i(x)
  {
  }
};

//*****************************************************************************
//
// TestOneToOneNoProperties
//
// DESCRIPTION:
//
//      Tests one-to-one associations that do not have properties.
//
//*****************************************************************************

void TestOneToOneNoProperties()
{
  A a;
  B b;
  B c;

  a.i = 1;
  b.i = 2;
  c.i = 4;

  ASToOne< A, B > a_link(a); // A's link
	ASToOne< B, A > b_link(b); // b's link
	ASToOne< B, A > c_link(c); // c's link

  // Link A and B together
  a_link.Attach(b_link);

  printf( "A is linked to %d\n", a_link.GetRhsObject().i );
  printf( "B is linked to %d\n", b_link.GetRhsObject().i );

  // Link A and C together
  a_link.Attach(c_link);

  printf( "A is linked to %d\n", a_link.GetRhsObject().i );
  printf( "C is linked to %d\n", c_link.GetRhsObject().i );
  
  // B should not be linked
  if ( b_link.IsAttached() )
  {
    printf( "B should not be linked to anything, but it's linked to %d\n", 
				b_link.GetRhsObject().i
          );
  }

  // Link a to itself using another association
  ASToOne< A, A > a_a_link(a), a_a_link2(a);
  a_a_link.Attach( a_a_link2 );

  printf( "A2 is linked to %d\n", a_a_link.GetRhsObject().i );
  
  // C and A should be attached at this point.
  c_link.DetachMe(); 
  if ( a_link.IsAttached() )
  {
    printf("Detach failed\n");
  }
  if ( c_link.IsAttached() )
  {
    printf("Detach failed\n");
  }
  
  // Create a link that will go out of scope,
  // resulting in a_link no longer being connected to anything.
    
  {
	 ASToOne< B,A > scoped_link(b);
	 a_link.Attach( scoped_link );
    
	 printf( "A is linked to %d\n", a_link.GetRhsObject().i );
  }
  
  if ( a_link.IsAttached() )
  {
    printf("Implicit detach failed\n");
  }
  
  // Test implicit "detach"
  ASToOne< A, B > a2_link(a); // A's link
  
  b_link.Attach( a_link );
  c_link.Attach( a2_link );
  b_link.Attach( a2_link );
  if ( c_link.IsAttached() )
  {
    printf("error 124\n");
  }
}

//*****************************************************************************
//
// TestOneToOneProp
//
// DESCRIPTION:
//
//      Tests one-to-one associations that have properties.
//
//*****************************************************************************

void TestOneToOneProp()
{
  A a;
  B b;

  a.i = 1;
  b.i = 2;

  ASToOneProp< A, B, Properties > a_link(a); // A's link
	ASToOneProp< B, A, Properties > b_link(b); // b's link

  // Link A and B together
  a_link.Attach(b_link, *new Properties(1));

  printf( "A is linked to %d\n", a_link.GetRhsObject().i );
  printf( "A properties %d\n", a_link.GetProperties().i );
  printf( "B is linked to %d\n", b_link.GetRhsObject().i );
  printf( "B properties %d\n", b_link.GetProperties().i );
  
}

//*****************************************************************************
//
// Typedefs for list and iterator 
//
//*****************************************************************************

typedef AS_LIST(ListTempl, A, B) List;
typedef AS_ITER_ADAPTER(A, B) Iterator;
typedef AS_LIST_PROP(ListTempl, A, B, Properties ) ListProp;
typedef AS_ITER_ADAPTER_PROP( A, B, Properties ) IteratorProp;
typedef AS_LIST( ListTempl, B, A ) BList;
typedef AS_ITER_ADAPTER( B, A ) BIterator;

//*****************************************************************************
//
// TestOneToManyNoProperties
//
// DESCRIPTION:
//
//      Tests one-to-many associations that do not have properties.
//
//*****************************************************************************

void TestOneToManyNoProperties()
{
  A a;
  B b;
  B c;

  a.i = 1;
  b.i = 2;
  c.i = 3;

  ASToOne< B, A > b_link(b); // B's link
  ASToOne< B, A > c_link(c); // C's link
  
  ASToMany< A, B > a_link(a, *new List); // A's link
  
  a_link.Attach( b_link );
  c_link.Attach( a_link ); 
  
  printf( "B is linked to %d\n", b_link.GetRhsObject().i );
  printf( "C is linked to %d\n", c_link.GetRhsObject().i );
  
  ASIteratorAdapter< ASAssn< B, A, ASNoProperties >
                   > *pIterator = a_link.GetRhsObjects().Iterate(); 
  
  while ( !pIterator -> AtEnd() )
  {
    printf( "Many link:  %d\n", pIterator->GetCurrent().GetLhsObject().i );
    pIterator -> Next();
  }
  
  delete pIterator;
}

//*****************************************************************************
//
// TestOneToManyProp
//
// DESCRIPTION:
//
//      Tests one-to-many associations that have properties.
//
//*****************************************************************************

void TestOneToManyProp()
{
  A a;
  B b;
  B c;

  a.i = 1;
  b.i = 2;
  c.i = 3;

	ASToOneProp< B, A, Properties > b_link(b); // b's link
  ASToOneProp< B, A, Properties > c_link(c); // c's link

  {
    ASToManyProp< A, B, Properties > a_link(a, *new ListProp); // A's link

    a_link.Attach( b_link, *new Properties(1) );
    c_link.Attach( a_link, *new Properties(2) ); 

    a_link.Detach( b_link );
  
    if ( b_link.IsAttached() )
      printf( "B is linked to %d\n", b_link.GetRhsObject().i );

    if ( c_link.IsAttached() )
      printf( "C is linked to %d\n", c_link.GetRhsObject().i );
  
    IteratorProp *pIterator = a_link.GetRhsObjects().Iterate(); 
  
    while ( !pIterator -> AtEnd() )
    {
      printf( "Many link:  %d\n", pIterator->GetCurrent().GetAssn().GetLhsObject().i );
      printf( "Props:  %d\n", pIterator->GetCurrent().GetProperties().i );
      pIterator -> Next();
    }
  
    delete pIterator;
  }

  // b and c should no longer be linked.
  if ( b_link.IsAttached() )
    printf( "B is linked to %d\n", b_link.GetRhsObject().i );
  if ( c_link.IsAttached() )
    printf( "C is linked to %d\n", c_link.GetRhsObject().i );
}

//*****************************************************************************
//
// TestManyToManyNoProperties
//
// DESCRIPTION:
//
//      Tests many-to-many associations that do not have properties.
//
//*****************************************************************************

void TestManyToManyNoProperties()
{
  A a;
  B b;
  B c;
  A d;

  a.i = 1;
  b.i = 2;
  c.i = 3;
  d.i = 4;

  ASToMany< B, A > b_link(b, *new BList ); // B's link (one to many)

  ASToOne< B, A > c_link(c); // C's link
  ASToOne< A, B > d_link(d); // D's link
  
  ASToMany< A, B > a_link(a, *new List); // A's link
  
  a_link.Attach( b_link );
  c_link.Attach( a_link ); 

  b_link.Attach(d_link);
  
  printf( "C is linked to %d\n", c_link.GetRhsObject().i );
  
  {
    ASIteratorAdapter< ASAssn< B, A, ASNoProperties >
                     > *pIterator = a_link.GetRhsObjects().Iterate(); 
    
    while ( !pIterator -> AtEnd() )
    {
      printf( "A Many link:  %d\n", pIterator->GetCurrent().GetLhsObject().i );
      pIterator -> Next();
    }
    
    delete pIterator;
  }

  ASIteratorAdapter< ASAssn< A, B, ASNoProperties >
                   > *pIterator = b_link.GetRhsObjects().Iterate(); 
  
  while ( !pIterator -> AtEnd() )
  {
    printf( "B Many link:  %d\n", pIterator->GetCurrent().GetLhsObject().i );
    pIterator -> Next();
  }
  
  delete pIterator;
}

//*****************************************************************************
// Main
//*****************************************************************************

#if 0

Expected output:

OneToOneNoProp
A is linked to 2
B is linked to 1
A is linked to 4
C is linked to 1
A2 is linked to 1
A is linked to 2

OneToOneProp
A is linked to 2
A properties 1
B is linked to 1
B properties 1

OneToManyNoProp
B is linked to 1
C is linked to 1
Many link:  2
Many link:  3

OneToManyProp
C is linked to 1
Many link:  3
Props:  2

ManyToMany
C is linked to 1
A Many link:  2
A Many link:  3
B Many link:  1
B Many link:  4

#endif


// Main
void main()
{
  printf( "OneToOneNoProp\n" );
  TestOneToOneNoProperties();
  printf( "\nOneToOneProp\n" );
  TestOneToOneProp();
  printf( "\nOneToManyNoProp\n" );
  TestOneToManyNoProperties();
  printf( "\nOneToManyProp\n" );
  TestOneToManyProp();
  printf( "\nManyToMany\n" );
  TestManyToManyNoProperties();
}
